# This file is part of TMart.
#
# Copyright 2024 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Unzip a .SAFE.zip file and return the path to the extracted .SAFE dir
# With contribution from Shun Bi

def unzip(zip_file_path):
    import zipfile
    unzip_folder = zip_file_path.replace('.SAFE.zip', '.SAFE')
    with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
        zip_ref.extractall(unzip_folder)
    print(f"Unzipped {zip_file_path} to {unzip_folder}")
    return unzip_folder



# This file is part of TMart.
#
# Copyright 2024 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Download hourly ancillary data from NASA Ocean Color 
def anci_download(file, basefile, username, password):
    
    import requests, time, os, sys
    url = 'https://oceandata.sci.gsfc.nasa.gov/ob/getfile/' + basefile
    auth = (username, password)
    local_file = file + '/tmart_ancillary/' + basefile
    
    if os.path.exists(local_file):
        print('File exists: ' + local_file)
        pass
    
    else:
        print('Downloading: ' + local_file)
        try:     
            if os.path.exists(os.path.dirname(local_file)) is False:
                os.makedirs(os.path.dirname(local_file))
            
            # source: https://github.com/acolite/acolite/blob/main/acolite/shared/download_file.py
            with requests.Session() as session:
                r1 = session.request('get', url, verify=True)
                r = session.get(r1.url, auth=auth, verify=True)
                time.sleep(1)
            
                if (r.ok):
                    with open(local_file, 'wb') as f:
                        for chunk in r.iter_content(chunk_size=1024*1024):
                            if chunk: f.write(chunk)
                  
        except: 
            sys.exit('Warning: failed to download: ' + str(basefile) + '. Please check your credentials and if file is available on NASA Ocean Color website. Note that the files are typically available a month after image acquisition.')
         
    return local_file                       
    
    
    
    
    
    
    
    
    
    
    
    
    
    







# This file is part of TMart.
#
# Copyright 2024 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Identify the type of the input: directory, file, or .SAFE.zip
# With contribution from Shun Bi

def identify_input(file_path):
    import os, sys
    import tmart
    
    # Initialize file_is_dir
    file_is_dir = None
    
    # If input is a directory
    if os.path.isdir(file_path):
        file_is_dir = True
        
    # If input is a file 
    elif os.path.isfile(file_path):
        # ACOLITE L1R (currently supporting PRISMA data only)
        if file_path.endswith('.L1R.nc'):
            file_is_dir = False
        # SAFE.zip file: unzip and update the path 
        elif file_path.endswith('.SAFE.zip'):
            file_path = tmart.AEC.unzip(file_path)
            file_is_dir = True
        
    # If the file or directory was not identified
    if file_is_dir is None:
        sys.exit(f"File or folder cannot be identified: {file_path}")
        
    return file_path, file_is_dir



# This file is part of TMart.
#
# Copyright 2024 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Overall control of AEC

def run(file, username, password, overwrite=False, AOT='MERRA2', n_photon=100_000, AOT_offset=0.0, n_jobs=100):
    '''Run adjacency-effect correction on satellite files. See 'Introduction - Adjacency-Effect Correction' for detailed instructions.
    
    Arguments:
        
    * ``file`` -- String. Path to satellite files. For L8/S2: provide path to the folder. For PRISMA: provide ACOLITE L1R file. SAFE.zip for S2 is supported: when doing so it is recommended to set ``overwrite`` as True to save space. 
    * ``username`` -- String. Username of EarthData account.
    * ``password`` -- String. Password of EarthData account.
    * ``overwrite`` -- Boolean. If overwrite the existing files. The default is False and it creates a folder in the same directory that starts with AEC in the name
    * ``AOT`` -- 'MERRA2': use ancillary data (default). Float: AOT at 550nm. 'NIR': calculate in T-Mart by finding dark pixels in NIR when considering the AE. 
    * ``n_photon`` -- Int. Number of photons in each T-Mart run, 100_000 is recommended for accurate results.
    * ``AOT_offset`` -- Float. Value added to AOT at 550nm. If resulted AOT is negative, it will be corrected to 0.
    * ``n_jobs`` -- Int. Number of jobs in Python multiprocessing. One CPU core processes one job at a time. n_photon is evenly distributed across the jobs.

    Example usage::
        
        import tmart
        file = 'user/test/S2A_MSIL1C_20160812T143752_N0204_R096_T20MKB_20160812T143749.SAFE'
        username = 'abcdef'
        password = '123456'
        
        # T-Mart uses multiprocessing, which needs to be wrapped in 'if __name__ == "__main__":' for Windows users
        if __name__ == "__main__":
            tmart.AEC.run(file, username, password)

    '''
    
    import tmart
    import sys, os
    
    import time
    from importlib.metadata import version
    
    # Identify directories and files 
    file, file_is_dir = tmart.AEC.identify_input(file)
    home_folder = os.path.dirname(file) # where ancillary file and printed info are stored, default: the dir of the file
    basename = os.path.basename(file)
    basename_before_period = basename.split('.')[0]
    
    # If not overwrite, make a copy of the file(s) in the same directory  
    if not overwrite:
        basename_new = 'AEC_' + basename
        file_new = os.path.join(home_folder,basename_new)
        if os.path.exists(file_new):
            print('\nNew directory ' + str(file_new) + ' already exists, proceeds to AEC. \n')
        else: 
            # make the copy 
            print('\nCopying from ' + str(file) + ' to ' + str(file_new) + '... \n')
            if file_is_dir: 
                from shutil import copytree
                copytree(file, file_new)
            else:
                from shutil import copy
                copy(file, file_new)
        file = file_new
    
    # For S2/L89, ancillary file and printed info are stored in the image folder 
    if file_is_dir: home_folder = file

    # Check if AEC has been done, if so, exit 
    AEC_record = '{}/AEC_completed_{}.txt'.format(home_folder,basename_before_period)
    if os.path.exists(AEC_record): 
        sys.exit('Record of adjacency correction found: ' + str(AEC_record) + 
                 '. Corrcting for the AE twice is not recommended and will lead to unpredictable results. ' + 
                 'If you’re confident that the imagery has not been corrected for the AE, delete this file to proceed.')

    # Start logging in txt file
    orig_stdout = sys.stdout
    log_file = 'tmart_log_' + time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime(time.time())) + '.txt'
    log_file = '{}/{}'.format(home_folder,log_file)

    class Logger:
        def __init__(self, filename):
            self.console = sys.stdout
            self.file = open(filename, 'w')
            self.file.flush()
        def write(self, message):
            self.console.write(message)
            self.file.write(message)
        def flush(self):
            self.console.flush()
            self.file.flush()

    sys.stdout = Logger(log_file)
    print('T-Mart version: ' + str(version('tmart')))
    print('System time: ' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time())))
    print('File: ' + str(file))
    print('AOT_offset: ' + str(AOT_offset))
    
    # S2/L89
    if file_is_dir: 
        tmart_out = tmart.AEC.run_regular(file, username, password, AOT, AOT_offset, n_photon, AEC_record, basename_before_period, n_jobs)
    # ACOLITE L1R 
    else:
        tmart_out = tmart.AEC.run_acoliteL1R(file, username, password, AOT, AOT_offset, n_photon, AEC_record, basename_before_period, n_jobs)
    
    # Stop logging 
    sys.stdout = orig_stdout

    return tmart_out
    


# This file is part of TMart.
#
# Copyright 2024 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Run on Landsat and Sentinel series 

def run_regular(file, username, password, AOT, AOT_offset, n_photon, AEC_record, basename, njobs): 
 
    import tmart
    import sys, os
    
    # read configuration
    config = tmart.AEC.read_config()

    # identify sensor
    print('\nReading image files: ')
    print(file)
    sensor = tmart.AEC.identify_sensor(file)
    
    # extract metadata
    if sensor == 'S2A' or sensor == 'S2B':
        metadata = tmart.AEC.read_metadata_S2(file, config, sensor)
    elif sensor == 'L8' or sensor == 'L9':
        metadata = tmart.AEC.read_metadata_Landsat(file, config, sensor)
    else: sys.exit('Warning: unrecognized sensor')
    
    metadata['sensor'] = sensor
    
    # print metadata 
    print('Metadata: ')
    for k, v in metadata.items():
        print(str(k) + ': '  + str(v))
        
    # get ancillary information from NASA Ocean Color
    anci = tmart.AEC.get_ancillary(metadata, username, password)
    
    # compute cloud and non-Water masks 
    print('\nComputing masks: ')
    mask_cloud = tmart.AEC.compute_masks(metadata, config, 'cloud')
    mask_all   = tmart.AEC.compute_masks(metadata, config, 'all')   
    print('Done')
    
    # AOT
    if AOT == 'NIR':
        print('\nEstimating AOT from the NIR band: ')
        AOT = tmart.AEC.get_AOT(metadata, config, anci, mask_cloud, mask_all, n_photon)
    elif AOT == 'MERRA2':
        AOT = anci['AOT_MERRA2']
        print('\nUsing AOT550 from MERRA2: ' + str(AOT))
    else:
        print('\nUser input AOT: ' + str(AOT))
    
    # add offset
    AOT = max(0, AOT + AOT_offset)
    
    # write atm information
    tmart.AEC.write_atm_info(file, basename, anci, AOT)
    print('\nWrote aerosol and atmosphere information.')
    
    # make a record file for AEC 
    file_AEC_record = open(AEC_record,"w")
    file_AEC_record.flush()
    
    # AEC for each of the specified bands 
    for i in range(len(metadata['AEC_bands_name'])):
        AEC_band_name = metadata['AEC_bands_name'][i] 
        AEC_band_6S = metadata['AEC_bands_6S'][i]
        wl = metadata['AEC_bands_wl'][i]
        print('\n============= AEC: {} ==================='.format(AEC_band_name))
        tmart.AEC.AEC(AEC_band_name, AEC_band_6S, wl, AOT, metadata, config, anci, mask_cloud, mask_all, n_photon, njobs)
        file_AEC_record.write(str(AEC_band_name) + '\n')
        file_AEC_record.flush()
        
    # close AEC record 
    file_AEC_record.close()
    
    return 0
    


# This file is part of T-Mart.
#
# Copyright 2024 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Compute total gas transmittance 

from Py6S import *
import numpy as np

def normal_distribution(x, mu, sigma):
    return 1 / (sigma * np.sqrt(2 * np.pi)) * np.exp(-((x - mu)**2) / (2 * sigma**2))

def calculate_heights(center_wavelength, FWHM, range_width=100, interval=2.5):
    sigma = FWHM / (2 * np.sqrt(2 * np.log(2)))
    start_wavelength = center_wavelength - range_width / 2
    end_wavelength = center_wavelength + range_width / 2
    wavelengths = np.arange(start_wavelength, end_wavelength + interval, interval)
    heights = normal_distribution(wavelengths, center_wavelength, sigma)
    
    # normalize heights 
    heights = heights / np.max(heights)

    return wavelengths[0], wavelengths[-1], heights

def compute_gas_transmittance(metadata, anci, wavelength, FWHM):

    s = SixS()
    s.geometry = Geometry.User()
    s.geometry.solar_z = metadata['sza']
    s.geometry.solar_a = 0
    s.geometry.view_z = 0 
    s.geometry.view_a = 0
    s.atmos_profile = AtmosProfile.UserWaterAndOzone(anci['water_vapour']/10, anci['ozone']/1000)
    s.altitudes.set_sensor_satellite_level()
    wl_start, wl_end, heights = calculate_heights(wavelength,FWHM)
    s.wavelength = Wavelength(wl_start/1000, wl_end/1000, heights.tolist())
    
    s.run()

    return s.outputs.transmittance_global_gas.total



# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

import numpy as np 
import pandas as pd
import sys
from copy import copy

# Analyze the output of TMart and differentiate direct, env and atm intrinsic reflectances  
def calc_ref(df, n_photon = None, detail = False):
    '''Analyze the results of T-Mart and calculate reflectances. 
    
    Arguments:

    * ``df`` -- Results from T-Mart runs
    * ``n_photon`` -- Specify the number of photons in the run when firing the photon upwards. If not specified, the number of unique pt_id will be used. This can lead to errors when photons were fired upwards because some photons will not have pt_id.
    * ``detail`` -- Boolean. Differentiate Cox-Munk, whitecap, water-leaving and land contributions

    Output:

    * A list of atmospheric intrinsic reflectance, direct reflectance, environmental reflectance and total reflectance. 

    Example usage::

      R = tmart.calc_ref(results)
      for k, v in R.items():
          print(k, '\t ' , v)

    '''
    
    print('=====================================')
    print('Calculating radiometric quantities...')
    
    # Columes: 0 pt_id, 1 movement, 2 L_cox-munk, 3 L_whitecap, 4 L_water, 5 L_land, 
    # 6 L_rayleigh, 7 L_mie, 8 9 10 surface xyz, 11 shadowed, 12 if_env
    
    if n_photon == None:
        n_photon = np.unique(df[:,0]).shape[0]
    
    R_atm = np.sum(df[:,6:8]) / n_photon
    R_dir = np.sum(df[df[:,12] == 0 ,2:6]) / n_photon # if_env == 1 and all surface reflectance 
    R_env = np.sum(df[df[:,12] == 1 ,2:6]) / n_photon # if_env == 0 and all surface reflectance 
    R_total = R_atm + R_dir + R_env
    
    if detail:
        R_dir_coxmunk   = np.sum(df[df[:,12] == 0 ,2]) / n_photon
        R_dir_whitecap  = np.sum(df[df[:,12] == 0 ,3]) / n_photon
        R_dir_water     = np.sum(df[df[:,12] == 0 ,4]) / n_photon
        R_dir_land      = np.sum(df[df[:,12] == 0 ,5]) / n_photon
        
        R_env_coxmunk   = np.sum(df[df[:,12] == 1 ,2]) / n_photon
        R_env_whitecap  = np.sum(df[df[:,12] == 1 ,3]) / n_photon
        R_env_water     = np.sum(df[df[:,12] == 1 ,4]) / n_photon
        R_env_land      = np.sum(df[df[:,12] == 1 ,5]) / n_photon        
        
        R_output = {'R_atm':R_atm,
                    'R_dir':R_dir,
                    '_R_dir_coxmunk':R_dir_coxmunk,
                    '_R_dir_whitecap':R_dir_whitecap,
                    '_R_dir_water':R_dir_water,
                    '_R_dir_land':R_dir_land,
                    'R_env':R_env,
                    '_R_env_coxmunk':R_env_coxmunk,
                    '_R_env_whitecap':R_env_whitecap,
                    '_R_env_water':R_env_water,
                    '_R_env_land':R_env_land,
                    'R_total':R_total}
    
    else:      
        R_output = {'R_atm':R_atm,
                    'R_dir':R_dir,
                    'R_env':R_env,
                    'R_total':R_total}
    
    
    return R_output



# This file is part of TMart.
#
# Copyright 2024 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Run on ACOLITE L1R files, currently supports PRSIMA only 

def run_acoliteL1R(file, username, password, AOT, AOT_offset, n_photon, AEC_record, basename, njobs):
 
    import tmart
    import sys, os, time
    import netCDF4 as nc4
    import numpy as np
    import scipy
        
    # Read configuration
    config = tmart.AEC.read_config()
    
    # Open netcdf file 
    dset = nc4.Dataset(file, 'r+')
    
    # wavelengths and widths 
    WLs = dset.getncattr('band_waves')
    WWs = dset.getncattr('band_widths')
    
    ### Set up metadata  
    window_size = 101
    reshape_factor = 5

    metadata = {}
    
    # Dimension 
    metadata['height'] = 1000
    metadata['width'] = 1000

    # Reshaped dimension for AEC
    height_reshaped = int(metadata['height'] / reshape_factor)
    width_reshaped = int(metadata['width'] / reshape_factor)
    
    # metadata and time 
    metadata['file'] = os.path.dirname(file)
    metadata['time'] = dset.getncattr('isodate')[0:19]
    
    metadata['sza'] = dset.getncattr('sza')
    metadata['saa'] = dset.getncattr('saa')
    metadata['vza'] = 0 # dset.getncattr('vza') # missing vaa for now so we assume 0 
    metadata['vaa'] = 0 # dset.getncattr('vaa')
    
    # find mean coordinates and direction to north 
    metadata['lat'] = np.mean(dset['lat'][:])
    metadata['lon'] = np.mean(dset['lon'][:])
    
    # Up relative to north 
    tm_PRISMA_up = tmart.AEC.read_PRISMA_north(file)
    sun_dir = [metadata['sza'], (metadata['saa']+270 - tm_PRISMA_up)%360]
        
    # Print metadata 
    print('Metadata: ')
    for k, v in metadata.items():
        print(str(k) + ': '  + str(v))
    
    # Get ancillary information from NASA Ocean Color
    anci = tmart.AEC.get_ancillary(metadata, username, password)
    
    
    ### Masks 
    
    # Cloud mask, may need to find band closest to 1373 in case of future wavelength shift 
    image_cloud = dset['rhot_1373'][:]
    image_cloud_compressed = image_cloud.reshape([height_reshaped, reshape_factor, width_reshaped, reshape_factor]).mean(3).mean(1)
    mask_cloud_compressed = image_cloud_compressed > float(config['mask_cirrus_threshold'])
    
    # SWIR, may need to find band closest to 1600 in case of future wavelength shift 
    mask_SWIR = dset['rhot_1596'][:] > float(config['mask_SWIR_threshold'])
    
    # TOA out of limit, all bands > 0.3 
    mask_highTOA = np.zeros(np.shape(mask_SWIR))    
    for WL in WLs: 
        print('High TOA: {:.1f} nm'.format(WL))
        rhot_wl = 'rhot_' + str(int(round(WL)))
        image_highTOA = dset[rhot_wl][:]
        if np.ma.count_masked(image_highTOA) > 0: 
            sys.exit('T-Mart assumed PRISMA has no masked values, please use another image or contact Yulun at yulunwu8@gmail.com')
        mask_highTOA = np.logical_or(mask_highTOA, image_highTOA > float(config['mask_highTOA_threshold']))
    
    # Combine highTOA and SWIR masks, original resolution 
    mask_all = np.logical_or(mask_highTOA, mask_SWIR)
    
    
    # AOT
    if AOT == 'NIR':
        sys.exit('The default estimating AOT from the NIR band is not tested on PRISMA data, please use MERRA2 or user input instead.')
    elif AOT == 'MERRA2':
        AOT = anci['AOT_MERRA2']
        print('\nUsing AOT from MERRA2: ' + str(AOT))
    else:
        print('\nUser input AOT: ' + str(AOT))
    
    # Add offset
    AOT = max(0, AOT + AOT_offset)
    
    # Write atm information
    tmart.AEC.write_atm_info(metadata['file'], basename, anci, AOT)
    print('\nWrote aerosol and atmosphere information.')
    
    
    ### T-Mart calculate correction parameters 

    # Loop through 5 wavelengths, generating 4 look-up tables    
    tm_wls_interp = np.array([406.9,550,783,1000,2429])
    list_conv_window_1  = []
    list_F_correction   = []
    list_F_captured     = []
    list_R_atm          = []
        
    for tm_wl_interp in tm_wls_interp: 
        wl = tm_wl_interp
        AEC_parameters = tmart.AEC.get_parameters(n_photon = n_photon, wl = wl, 
                                                  target_pt_direction=[180,0], sun_dir=sun_dir, 
                                                  atm_profile = anci, 
                                                  aerosol_type = anci['r_maritime'], aot550 = AOT, 
                                                  cell_size = 30,
                                                  window_size = window_size, njobs=njobs)
        
        conv_window_1   = AEC_parameters['conv_window_1']
        F_correction    = AEC_parameters['F_correction']
        F_captured      = AEC_parameters['F_captured']
        R_atm           = AEC_parameters['R_atm']
    
        list_conv_window_1.append(conv_window_1)
        list_F_correction.append(F_correction)
        list_F_captured.append(F_captured)
        list_R_atm.append(R_atm)
    
    # make np arrays 
    list_conv_window_1 = np.array(list_conv_window_1)
    list_F_correction = np.array(list_F_correction)
    list_F_captured = np.array(list_F_captured)
    list_R_atm = np.array(list_R_atm)
    
    # interpolation objects  
    interp_conv_window_1 = scipy.interpolate.interp1d(tm_wls_interp, list_conv_window_1, kind='linear', axis=0)
    interp_F_correction = scipy.interpolate.interp1d(tm_wls_interp, list_F_correction, kind='linear', axis=0)
    interp_F_captured = scipy.interpolate.interp1d(tm_wls_interp, list_F_captured, kind='linear', axis=0)
    interp_R_atm = scipy.interpolate.interp1d(tm_wls_interp, list_R_atm, kind='linear', axis=0)
        
    # Make a record file for AEC 
    file_AEC_record = open(AEC_record,"w")
    file_AEC_record.flush()
    
    # Loop through all bands 
    for i in range(len(WLs)):

        # Wavelength and FWHM
        WL = WLs[i]
        WW = WWs[i]
        
        print('\n----------')
        print('Processing wavelength {:.2f} nm'.format(WL))
        
        # If tgas < 0.7, then skip AEC for this wavelength 
        tgas = tmart.AEC.compute_gas_transmittance(metadata, anci, WL, WW)
        if tgas < 0.7: 
            print('Skipping, tgas: {:.2f} < 0.7'.format(tgas))
            continue
        else:
            print('tgas: {:.2f}'.format(tgas))
            
        # Skip if out of range 
        if WL > tm_wls_interp.max():
            print('Skipping band out of correction-parameter range')
            continue
    
        # Variable name 
        rhot_wl = 'rhot_' + str(int(round(WL)))
        
        # interpolate AEC parameters 
        conv_window_1   = interp_conv_window_1(WL)
        F_correction    = interp_F_correction(WL)
        F_captured      = interp_F_captured(WL)
        R_atm           = interp_R_atm(wl)
        
        # Read image 
        image = dset[rhot_wl][:].data
        
        # Turn negative TOA to 0 
        image[image<0] = 0
        
        # Reshape
        image_AEC = image.reshape([height_reshaped, reshape_factor, 
                                   width_reshaped, reshape_factor]).mean(3).mean(1)
        
        # image_R_surf: reshaped 
        image_R_surf = image_AEC - R_atm
        print('\nNumber of negative R_surf pixels: ' + str(np.sum(image_R_surf < 0)) + '/' + str(height_reshaped * width_reshaped))
        
        # Turn negative to 0
        image_R_surf[image_R_surf<0] = 0
        
        # Read AEC parameters 
        hp_cloud_contribution = float(config['cloud_contribution'])
        hp_AE_land = config['AE_land'] == 'True'
        
        # Reduce contribution of clouds according to setting 
        image_R_surf[mask_cloud_compressed] = image_R_surf[mask_cloud_compressed] * hp_cloud_contribution
        
        # Convolution 
        start_time = time.time()
        print("\nConvolution started ")
        filter_kernel = np.flip(conv_window_1) # it's flipped in convolve by default
        R_conv = scipy.signal.convolve2d(image_R_surf, filter_kernel, mode='same', boundary='fill', fillvalue=image_R_surf.mean())
        print("Convolution completed: %s seconds " % (time.time() - start_time))
        
        
        # Smoothing the edges 
        if reshape_factor >1 and not hp_AE_land:
            image_water = np.where(mask_all, np.nan, image)
            
            # Suppress warning of mean of empty slice 
            import warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                image_water_AEC = image_water.reshape([height_reshaped, reshape_factor, width_reshaped, reshape_factor])
                image_water_AEC = np.nanmean(image_water_AEC,3)
                image_water_AEC = np.nanmean(image_water_AEC,1)
        
            # with non-water as nan 
            image_water_R_surf = image_water_AEC - R_atm
            
            # Edit image_R_surf: where there's water image_R_surf, keep it, where there's nan, use original image_R_surf
            # image_R_surf can only be edited after convolution 
            image_R_surf = np.where(np.isnan(image_water_R_surf), image_R_surf, image_water_R_surf)   
            
        # Correction to make 
        R_correction = (R_conv - image_R_surf) * F_correction 
        print('\nNumber of pixels where R_correction > R_surf : ' + str(np.sum(R_correction>image_R_surf)) + '/' + str(height_reshaped * width_reshaped))
    
        # Back to the original size 
        R_correction_original_shape = np.repeat(np.repeat(R_correction, reshape_factor, axis=0), reshape_factor, axis=1)
    
        # Smoothing the gridline artifacts 
        if reshape_factor>1: 
        
            # Keep only water, short name for R_correction_original_shape_water
            RC_water = np.where(mask_all, np.nan, R_correction_original_shape)
            RC_water_fill_nan = tmart.AEC.fillnan(RC_water)
            
            filter_size = 3
            filter_kernel = np.full((filter_size,filter_size), 1/(filter_size**2))
            
            RC_water_smooth = scipy.signal.convolve2d(RC_water_fill_nan, filter_kernel, mode='same', boundary='fill', fillvalue=np.nanmean(RC_water))
            
            # place RC_water_smooth in R_correction_original_shape where there's water
            R_correction_original_shape = np.where(mask_all, R_correction_original_shape, RC_water_smooth)
    
        # remove AE from TOA reflectance 
        temp_SR = image - R_correction_original_shape - R_atm
        
        # Correcting for non-linearity of the ratio of environmental irradiance to surface reflectance for homogeneous Lambertian surfaces
        # This is implemented for water only
        # Input has to include land, because the function uses the env irradiance of the average reflectance across the scene
        print("\nIrradiance correction: ")
        temp_SR_water = tmart.AEC.irradiance_correction(image = temp_SR, wl_RC = wl/1000, band = None,
                                                        tm_vza = metadata['vza'], tm_vaa = metadata['vaa'], 
                                                        tm_sza = metadata['sza'], tm_saa = metadata['saa'],
                                                        atm_profile = anci, 
                                                        aerosol_type = anci['r_maritime'], aot550 = AOT)
        
        # If land correction is needed 
        if hp_AE_land:
            temp_out = np.where(mask_all, temp_SR, temp_SR_water)
        else:
            temp_out = np.where(mask_all, (image-R_atm), temp_SR_water)

        temp_out = temp_out + R_atm
        
        # Modify L1R NC file 
        dset[rhot_wl][:] = temp_out
            
        # Write AEC record
        file_AEC_record.write(str(rhot_wl) + '\n')
        file_AEC_record.flush()
    
    # Close AEC record 
    file_AEC_record.close()
    
    # Close file 
    dset.close()
    
    return 0
    
 